//                    kepler.cc
/////////////////////////////////////////////////////////////
// 04.06.2002: Runge-Kutta-Verfahren vierter Ordnung
//             rk4() implementiert

#include <iostream>
#include <fstream>
#include <cmath>

////////// Erst einige Definitionen

const double Pi = 3.14159265358979323846264;

const int rdim = 2;		// 2 raeumliche Dimensionen
const int nobj = 2;		// Die Anzahl der Objekte

double Gm[nobj] = {		// Die Produkte G m in AU
          4*Pi*Pi,		// Sonne
          0.000119		// Erde	       
       };

const int aktZahl = 1;		// Zahl der zu aktualisiernden
				// Koerper
int akt[aktZahl] = {1};	        // und deren Indizes

char *name[aktZahl] = {		// die "Datei"-Namen
   "erde.dat"};	                // dazu

double r0[nobj] = {0, 1};	// Bahnradien von Sonne, Erde

const double deltaT = 0.000013;	// Die Schrittweite in Jahren 
const double Tgesamt = 10.0*1.0;	// 10 Erd-Umlaeufe
//const double Tgesamt = 100000*deltaT;	// 10 Erd-Umlaeufe

////////// nun das eigentliche Programm

// Die Beschleunigung, die Koerper i durch die anderen erfaehrt

void beschl(int i, double r[nobj][rdim], double b[rdim])
 {
  double rij[rdim];
  double rabs, temp;

  for(int n=0; n<rdim; n++)	// b auf Null
     b[n] = 0;			// initialisieren

  for(int j=0; j<nobj; j++)	// Schleife ueber alle
    if(j != i)			// ANDEREN Objekte
      {
        temp = 0;		// temp auf Null initialisieren
        for(int n=0; n<rdim; n++)
          {
            rij[n] = r[i][n] - r[j][n];	// Differenzvektor bestimmen
            temp += (rij[n]*rij[n]);	// Quadrat in temp
          }
        rabs = sqrt(temp);	// Wurzel ziehen
        temp *= rabs;		// davon die 3. Potenz in temp

        for(int n=0; n<rdim; n++)	// Schliesslich zu b addieren
            b[n] -= (Gm[j]*rij[n])/temp;
      }
 }

// Ein Euler-Schritt mit deltaT

void euler(double r[nobj][rdim], double v[nobj][rdim], double deltaT)
 {
  double be[aktZahl][rdim];

  for(int i=0; i<aktZahl; i++)	// Zuerst alle benoetigten
    beschl(akt[i], r, be[i]);	// Beschleunigungen berechnen

  for(int i=0; i<aktZahl; i++)	// Dann die
   for(int n=0; n<rdim; n++)
    {
     r[akt[i]][n] += v[akt[i]][n]*deltaT;	// Orte und
     v[akt[i]][n] += be[i][n]*deltaT;		// Geschwindigkeiten
    }				// aktualisieren
 }

// Ein Euler-Richardson-Schritt mit deltaT

void eulerRichardson(double r[nobj][rdim],
                     double v[nobj][rdim], double deltaT)
 {
  double be[aktZahl][rdim], bmid[aktZahl][rdim];
  double rmid[nobj][rdim], vmid[nobj][rdim];

  for(int i=0; i<aktZahl; i++)	// Erst einmal alle
    beschl(akt[i], r, be[i]);	// Beschleunigungen berechnen

  for(int i=0; i<nobj; i++)
   for(int n=0; n<rdim; n++)
    {				// Sicherheitshalber
     rmid[i][n] = r[i][n];	// r und
     vmid[i][n] = v[i][n];	// v in rmid und vmid
    }				// kopieren

  for(int i=0; i<aktZahl; i++)	// Dann ein Euler-Schritt
   for(int n=0; n<rdim; n++)	// mit deltaT/2
    {
     rmid[akt[i]][n] += v[akt[i]][n]*deltaT/2;
     vmid[akt[i]][n] += be[i][n]*deltaT/2;
    }

  for(int i=0; i<aktZahl; i++)	// Beschleunigungen noch
    beschl(akt[i], rmid, bmid[i]);	// einmal berechnen

  for(int i=0; i<aktZahl; i++)	// Und schliesslich
   for(int n=0; n<rdim; n++)	// der Schritt ueber deltaT 
    {				// unter Verwendung von
     r[akt[i]][n] += vmid[akt[i]][n]*deltaT;	// vmid
     v[akt[i]][n] += bmid[i][n]*deltaT;		// und bmid
    } 
 }

// Ein Predictor-Corrector Schritt mit deltaT

void predictor(double r[nobj][rdim],
                     double v[nobj][rdim], double deltaT)
 {
  double be[aktZahl][rdim], bpred[aktZahl][rdim];
  double rpred[nobj][rdim], vpred[nobj][rdim];

  for(int i=0; i<aktZahl; i++)	// Erst einmal alle
    beschl(akt[i], r, be[i]);	// Beschleunigungen berechnen

  for(int i=0; i<nobj; i++)
   for(int n=0; n<rdim; n++)
    {				// Sicherheitshalber
     rpred[i][n] = r[i][n];	// r und
     vpred[i][n] = v[i][n];	// v in rpred und vpred
    }				// kopieren

  for(int i=0; i<aktZahl; i++)	// Dann ein Euler-Schritt
   for(int n=0; n<rdim; n++)	// mit deltaT
    {
     rpred[akt[i]][n] += v[akt[i]][n]*deltaT;
     vpred[akt[i]][n] += be[i][n]*deltaT;
    }

  for(int i=0; i<aktZahl; i++)	// Beschleunigungen noch
    beschl(akt[i], rpred, bpred[i]);	// einmal berechnen

  for(int i=0; i<aktZahl; i++)	// Und schliesslich
   for(int n=0; n<rdim; n++)	// der Schritt ueber deltaT 
    {				// unter Verwendung von
     r[akt[i]][n] += v[akt[i]][n]*deltaT/2 + vpred[akt[i]][n]*deltaT/2;	// vpred
     v[akt[i]][n] += be[i][n]*deltaT/2     + bpred[i][n]*deltaT/2;	// und bpred
    } 
 }
 
// Runge-Kutta-Verfahren vierter Ordnung
// Schritt mit deltaT

const int kAnZahl = 4;			// Anzahl der ki

// Zaehler fuer Berechnung von k1, ..., k4
const int kfac[kAnZahl-1] = {2, 2, 1};
// Zaehler der Gewichtsfaktoren im Ergebnis
const int ksfac[kAnZahl] = {6, 3, 3, 6};

void rk4(double r[nobj][rdim], double v[nobj][rdim], double deltaT)
 {
  double b[aktZahl][rdim];		// Die Beschleunigungen
  double kr[kAnZahl][aktZahl][rdim];	// r- und v- Komponenten
  double kv[kAnZahl][aktZahl][rdim];	// von k1, ..., k4
  double rt[nobj][rdim], vt[nobj][rdim];

  for(int i=0; i<nobj; i++)
   for(int n=0; n<rdim; n++)
    {					// Sicherheitshalber
     rt[i][n] = r[i][n];		// r und v in
     vt[i][n] = v[i][n];		// rt und vt kopieren
    }

  for(int kzahl=0; kzahl<kAnZahl; kzahl++)	// k1, ..., k4 berechnen
   {
    for(int i=0; i<aktZahl; i++)	// Erst einmal alle
      beschl(akt[i], rt, b[i]);		// Beschleunigungen berechnen
    for(int i=0; i<aktZahl; i++)
     for(int n=0; n<rdim; n++)
      {
       kv[kzahl][i][n] = b[i][n]*deltaT;
       kr[kzahl][i][n] = vt[akt[i]][n]*deltaT;
      }
    if(kzahl < (kAnZahl-1))		// rt und vt fuer den
      for(int i=0; i<aktZahl; i++)	// naechsten Schritt
       for(int n=0; n<rdim; n++)	// vorbereiten
        {
         rt[akt[i]][n] = r[akt[i]][n] + kr[kzahl][i][n]/kfac[kzahl];
         vt[akt[i]][n] = v[akt[i]][n] + kv[kzahl][i][n]/kfac[kzahl];
        }
   }

  for(int i=0; i<aktZahl; i++)		// Und schliesslich k1, ..., k4
   for(int n=0; n<rdim; n++)		// mit den richtigen Faktoren
    for(int kzahl=0; kzahl<kAnZahl; kzahl++)	// zu r und v addieren
      {
       r[akt[i]][n] += kr[kzahl][i][n]/ksfac[kzahl];
       v[akt[i]][n] += kv[kzahl][i][n]/ksfac[kzahl];
      } 
 }

// Hauptprogramm

int main()
 {
  double r[nobj][rdim], v[nobj][rdim];
  std::ofstream ausgabe[aktZahl];

  for(int i=0; i<aktZahl; i++)	// Ausgabedateien oeffnen
    ausgabe[i].open(name[i], std::ios::out);

  for(int i=0; i<nobj; i++)	// Initialisierung
   {
    r[i][0] = r0[i];		// x-Koordinate von r
    r[i][1] = 0;		// y-Koordinate auf 0
    v[i][0] = 0;		// Anfangsgeschwindigkeit
    if(! i)			// so, dass es eine Kreis-
      v[i][1] = 0;		// bahn um die Sonne gibt
    else
      v[i][1] = sqrt(Gm[0]/r0[i]);
   }
  
  for(int i=0; i<aktZahl; i++)	        // Die Orginal-KOO schreiben:
      ausgabe[i] << 0			// Zeit t=0
      		 << "\t"		// sowie 
      		 << r[akt[i]][0]	// x-Koordinaten
                 << "\t"		// und 
                 << r[akt[i]][1]	// und y-Koordinaten
	         << "\n";		// ausgeben

		 
  for(double t=0; t<=Tgesamt; t += deltaT)
   {
       euler(r, v, deltaT);	        // Euler-Schritt

       //eulerRichardson(r, v, deltaT);	// Euler-Richardson-Schritt

       //predictor(r, v, deltaT);	// Predictor-Corrector-Schritt

       //rk4(r, v, deltaT);             // Runge-Kutta-Schritt

    for(int i=0; i<aktZahl; i++)	// Dann
      ausgabe[i] << t+deltaT			// Zeit t
    		 << "\t"		// sowie 
    		 << r[akt[i]][0]	// x- 
                 << "\t"		// und 
                 << r[akt[i]][1]	// y-Koordinaten
	         << "\n";		// ausgeben
   }

  for(int i=0; i<aktZahl; i++)
     ausgabe[i].close();	// Ausgabedateien schliessen
 }
