//                    kepler.cc
/////////////////////////////////////////////////////////////
// 04.06.2002: Runge-Kutta-Verfahren vierter Ordnung
//             rk4() implementiert

#include <iostream>
#include <fstream>
#include <cmath>

////////// Erst einige Definitionen
#define OUTPUT 20

const double Pi = 3.14159265358979323846264;

const double AU = 149597870691.0;
const double yr = 365*24*3600+6*3600+9*60+9.54;
const double g = 6.6742e-11;

const int rdim = 3;		// raeumliche Dimensionen
const int nobj = 10;		// Anzahl der Objekte

double Gm[nobj] = {		// products g*m in (AU^3 / yr^2)
          4*Pi*Pi,		// Sonne
	  g*(0.3302e+24)*yr*yr/(AU*AU*AU), // Merkur	       
	  g*(4.8685e+24)*yr*yr/(AU*AU*AU),  // Venus
          g*(5.9736e+24)*yr*yr/(AU*AU*AU), // Erde
	  g*(0.64185e+24)*yr*yr/(AU*AU*AU),  // Mars   
	  g*(1898.6e+24)*yr*yr/(AU*AU*AU), // Jupiter
	  g*(568.46e+24)*yr*yr/(AU*AU*AU), // Saturn
	  g*(86.832e+24)*yr*yr/(AU*AU*AU), // Uranus
	  g*(102.43e+24)*yr*yr/(AU*AU*AU), // Neptun
	  g*(0.0125e+24)*yr*yr/(AU*AU*AU), // Pluto
       };

const int aktZahl = 10;		// Zahl der zu aktualisiernden
				// Koerper
int akt[aktZahl] = {0,1,2,3,4,5,6,7,8,9};	        // und deren Indizes

char *name[aktZahl] = {		// die "Datei"-Namen
    "sonne.dat",
    "merkur.dat",
    "venus.dat",
    "erde.dat",
   "mars.dat",
    "jupiter.dat",
    "saturn.dat",
     "uranus.dat",
     "neptun.dat",
     "pluto.dat"};	              

// Mittlere Bahnelemente gelten fuer J2000 von 
// (Mittelwerte ber die letzten 250Jahre)
// http://nssdc.gsfc.nasa.gov/planetary/factsheet/index.html
double a[nobj] = {0.0, 0.38709893, 0.72333199, 1.00000011, 1.52366231, 5.20336301, 9.53707032, 19.19126393, 30.06896348, 39.48168677}; //grosse Halbachse 
double Ex[nobj] = {0.0, 0.20563069, 0.00677323, 0.01671022, 0.09341233, 0.04839266, 0.05415060, 0.04716771, 0.00858587, 0.24880766}; //Exzentrizitaet Pluto 0.24880766
//double Ex[nobj] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double In[nobj] = {0.0, 7.00487, 3.39471, 0.00005, 1.85061, 1.30530, 2.48446, 0.76986, 1.76917, 17.14175}; //Inklination
//double In[nobj] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double Wm[nobj] = {0.0, 48.33167, 76.68069, -11.26064, 49.57854, 100.55615, 113.71504, 74.22988, 131.72169, 110.30347}; //Knotenlaenge
//double Wm[nobj] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
double wl[nobj] = {0.0, 77.45645, 131.53298, 102.94719, 336.04084, 14.75385, 92.43194, 170.96424, 44.97135, 224.06676}; //Perihellnge
//double wl[nobj] = {0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0};
//double nu[nobj] =  {0.0, 77.45645, 131.53298, 102.94719, 336.04084, 14.75385, 92.43194, 170.96424, 44.97135, 224.06676}; //Perihellnge
double nu[nobj] = {0.0, 252.25084, 181.97973, 100.46435, 355.45332, 34.40438, 49.94432, 313.23218, 304.88003, 238.92881}; //mittlere Lnge
//The longitude a body was at in its orbit at 12:00 Universal (Greenwich) Time on January 1, 2000, also known as J2000 or Julian Day 2451545.0 (vom Frhlingspunkt)

const double deltaT = 0.001;	// Die Schrittweite 0.005

const double Tgesamt = 248.0*1.0;	// 10 Erd-Umlaeufe Pluto: (sqrt(a[9]*a[9]*a[9]))

////////// nun das eigentliche Programm

// Die Beschleunigung, die Koerper i durch die anderen erfaehrt

void beschl(int i, double r[nobj][rdim], double b[rdim])
 {
  double rij[rdim];
  double rabs, temp;

  for(int n=0; n<rdim; n++)	// b auf Null
     b[n] = 0;			// initialisieren

  for(int j=0; j<nobj; j++)	// Schleife ueber alle
    if(j != i)			// ANDEREN Objekte
      {
        temp = 0;		// temp auf Null initialisieren
        for(int n=0; n<rdim; n++)
          {
            rij[n] = r[i][n] - r[j][n];	// Differenzvektor bestimmen
            temp += (rij[n]*rij[n]);	// Quadrat in temp
          }
        rabs = sqrt(temp);	// Wurzel ziehen
        temp *= rabs;		// davon die 3. Potenz in temp

        for(int n=0; n<rdim; n++)	// Schliesslich zu b addieren
            b[n] -= (Gm[j]*rij[n])/temp;
      }
 }

// Ein Euler-Schritt mit deltaT

void euler(double r[nobj][rdim], double v[nobj][rdim], double deltaT)
 {
  double be[aktZahl][rdim];

  for(int i=0; i<aktZahl; i++)	// Zuerst alle benoetigten
    beschl(akt[i], r, be[i]);	// Beschleunigungen berechnen

  for(int i=0; i<aktZahl; i++)	// Dann die
   for(int n=0; n<rdim; n++)
    {
     r[akt[i]][n] += v[akt[i]][n]*deltaT;	// Orte und
     v[akt[i]][n] += be[i][n]*deltaT;		// Geschwindigkeiten
    }				// aktualisieren
 }

// Ein Euler-Richardson-Schritt mit deltaT

void eulerRichardson(double r[nobj][rdim],
                     double v[nobj][rdim], double deltaT)
 {
  double be[aktZahl][rdim], bmid[aktZahl][rdim];
  double rmid[nobj][rdim], vmid[nobj][rdim];

  for(int i=0; i<aktZahl; i++)	// Erst einmal alle
    beschl(akt[i], r, be[i]);	// Beschleunigungen berechnen

  for(int i=0; i<nobj; i++)
   for(int n=0; n<rdim; n++)
    {				// Sicherheitshalber
     rmid[i][n] = r[i][n];	// r und
     vmid[i][n] = v[i][n];	// v in rmid und vmid
    }				// kopieren

  for(int i=0; i<aktZahl; i++)	// Dann ein Euler-Schritt
   for(int n=0; n<rdim; n++)	// mit deltaT/2
    {
     rmid[akt[i]][n] += v[akt[i]][n]*deltaT/2;
     vmid[akt[i]][n] += be[i][n]*deltaT/2;
    }

  for(int i=0; i<aktZahl; i++)	// Beschleunigungen noch
    beschl(akt[i], rmid, bmid[i]);	// einmal berechnen

  for(int i=0; i<aktZahl; i++)	// Und schliesslich
   for(int n=0; n<rdim; n++)	// der Schritt ueber deltaT 
    {				// unter Verwendung von
     r[akt[i]][n] += vmid[akt[i]][n]*deltaT;	// vmid
     v[akt[i]][n] += bmid[i][n]*deltaT;		// und bmid
    } 
 }
 
// Runge-Kutta-Verfahren vierter Ordnung
// Schritt mit deltaT

const int kAnZahl = 4;			// Anzahl der ki

// Zaehler fuer Berechnung von k1, ..., k4
const int kfac[kAnZahl-1] = {2, 2, 1};
// Zaehler der Gewichtsfaktoren im Ergebnis
const int ksfac[kAnZahl] = {6, 3, 3, 6};

void rk4(double r[nobj][rdim], double v[nobj][rdim], double deltaT)
 {
  double b[aktZahl][rdim];		// Die Beschleunigungen
  double kr[kAnZahl][aktZahl][rdim];	// r- und v- Komponenten
  double kv[kAnZahl][aktZahl][rdim];	// von k1, ..., k4
  double rt[nobj][rdim], vt[nobj][rdim];

  for(int i=0; i<nobj; i++)
   for(int n=0; n<rdim; n++)
    {					// Sicherheitshalber
     rt[i][n] = r[i][n];		// r und v in
     vt[i][n] = v[i][n];		// rt und vt kopieren
    }

  for(int kzahl=0; kzahl<kAnZahl; kzahl++)	// k1, ..., k4 berechnen
   {
    for(int i=0; i<aktZahl; i++)	// Erst einmal alle
      beschl(akt[i], rt, b[i]);		// Beschleunigungen berechnen
    for(int i=0; i<aktZahl; i++)
     for(int n=0; n<rdim; n++)
      {
       kv[kzahl][i][n] = b[i][n]*deltaT;
       kr[kzahl][i][n] = vt[akt[i]][n]*deltaT;
      }
    if(kzahl < (kAnZahl-1))		// rt und vt fuer den
      for(int i=0; i<aktZahl; i++)	// naechsten Schritt
       for(int n=0; n<rdim; n++)	// vorbereiten
        {
         rt[akt[i]][n] = r[akt[i]][n] + kr[kzahl][i][n]/kfac[kzahl];
         vt[akt[i]][n] = v[akt[i]][n] + kv[kzahl][i][n]/kfac[kzahl];
        }
   }

  for(int i=0; i<aktZahl; i++)		// Und schliesslich k1, ..., k4
   for(int n=0; n<rdim; n++)		// mit den richtigen Faktoren
    for(int kzahl=0; kzahl<kAnZahl; kzahl++)	// zu r und v addieren
      {
       r[akt[i]][n] += kr[kzahl][i][n]/ksfac[kzahl];
       v[akt[i]][n] += kv[kzahl][i][n]/ksfac[kzahl];
      } 
 }

// Hauptprogramm

int main()
 {
  double r[nobj][rdim], rhelio[nobj], e[nobj][rdim], C[nobj][rdim], vperih[nobj], velip[nobj], v[nobj][rdim], Impx, Impy, Impz;
  long int steps;
  std::ofstream ausgabe[aktZahl];
  steps=0;

  Impx=0.0;
  Impy=0.0;
  Impz=0.0;
  for(int i=1; i<nobj; i++) {
      //Bahnvektor
      e[i][0] = cos(Wm[i]/180.0*Pi)*cos(((nu[i]-Wm[i]))/180.0*Pi)-sin(Wm[i]/180.0*Pi)*sin(((nu[i]-Wm[i]))/180.0*Pi)*cos(In[i]/180.0*Pi);
      e[i][1] = sin(Wm[i]/180.0*Pi)*cos(((nu[i]-Wm[i]))/180.0*Pi)+cos(Wm[i]/180.0*Pi)*sin(((nu[i]-Wm[i]))/180.0*Pi)*cos(In[i]/180.0*Pi);
      e[i][2] = sin(((nu[i]-Wm[i]))/180.0*Pi)*sin(In[i]/180.0*Pi);
      //Drehimpulsvektor
      C[i][0] =  sin(Wm[i]/180.0*Pi)*sin(In[i]/180.0*Pi);
      C[i][1] =  -cos(Wm[i]/180.0*Pi)*sin(In[i]/180.0*Pi);
      C[i][2] =  cos(In[i]/180.0*Pi); 
     rhelio[i] = a[i]*(1.0-Ex[i])*(1.0+Ex[i])/(1.0+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi));
     //velip[i] = sqrt(Gm[0]*((2.0/rhelio[i])-(1.0/a[i]))); //Geschw auf Ellipse
     //vperih[i] = sqrt(Gm[0]/a[i])*sqrt((1+Ex[i])/(1-Ex[i]));
     //v[i][0] = vperih[i]*(C[i][1]*e[i][2]-C[i][2]*e[i][1]);
     //v[i][1] = vperih[i]*(C[i][2]*e[i][0]-C[i][0]*e[i][2]);
     //v[i][2] = vperih[i]*(C[i][0]*e[i][1]-C[i][1]*e[i][0]);

     v[i][0] = sqrt(Gm[0]/(a[i]*(1-Ex[i]*Ex[i])))*(Ex[i]*sin((nu[i]-wl[i])/180.0*Pi)*e[i][0]+(1+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi))*(C[i][1]*e[i][2]-C[i][2]*e[i][1]));
     v[i][1] = sqrt(Gm[0]/(a[i]*(1-Ex[i]*Ex[i])))*(Ex[i]*sin((nu[i]-wl[i])/180.0*Pi)*e[i][1]+(1+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi))*(C[i][2]*e[i][0]-C[i][0]*e[i][2]));
     v[i][2] = sqrt(Gm[0]/(a[i]*(1-Ex[i]*Ex[i])))*(Ex[i]*sin((nu[i]-wl[i])/180.0*Pi)*e[i][2]+(1+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi))*(C[i][0]*e[i][1]-C[i][1]*e[i][0]));

      Impx+=Gm[i]*v[i][0];
      Impy+=Gm[i]*v[i][1];
      Impz+=Gm[i]*v[i][2];
  }
   
  for(int i=0; i<aktZahl; i++)	// Ausgabedateien oeffnen
    ausgabe[i].open(name[i], std::ios::out);

  for(int i=0; i<nobj; i++)	// Initialisierung
   {

      e[i][0] = cos(Wm[i]/180.0*Pi)*cos(((nu[i]-Wm[i]))/180.0*Pi)-sin(Wm[i]/180.0*Pi)*sin(((nu[i]-Wm[i]))/180.0*Pi)*cos(In[i]/180.0*Pi);
      e[i][1] = sin(Wm[i]/180.0*Pi)*cos(((nu[i]-Wm[i]))/180.0*Pi)+cos(Wm[i]/180.0*Pi)*sin(((nu[i]-Wm[i]))/180.0*Pi)*cos(In[i]/180.0*Pi);
      e[i][2] = sin(((nu[i]-Wm[i]))/180.0*Pi)*sin(In[i]/180.0*Pi);

      C[i][0] =  sin(Wm[i]/180.0*Pi)*sin(In[i]/180.0*Pi);
      C[i][1] =  -cos(Wm[i]/180.0*Pi)*sin(In[i]/180.0*Pi); 
      C[i][2] =  cos(In[i]/180.0*Pi); 
    rhelio[i] = a[i]*(1.0-Ex[i])*(1.0+Ex[i])/(1.0+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi));
     
    
      //v[i][0] = 0;		// Anfangsgeschwindigkeit
    if(! i) {			// so, dass es eine Kreis-
	//v[i][1] = 0;		// bahn um die Sonne gibt
	//v[i][1] = -2*Pi*0.005/9.2; //fast OK bei r0=0.005
	
	v[i][0] = -Impx/Gm[0];
	v[i][1] = -Impy/Gm[0];
        v[i][2] = -Impz/Gm[0];
	r[i][0] = 0.0;//-0.007;	
         r[i][1] = 0.0;//-0.003;	
         r[i][2] = 0.0;//0.0002;
    }
    else   {                     // (0 = falsch, 1 = wahr, !Bed. = Negation)
	
      r[i][0] = rhelio[i]* e[i][0];	
      r[i][1] = rhelio[i]* e[i][1];	
      r[i][2] = rhelio[i]* e[i][2]; 
         
      //velip[i]=sqrt(Gm[0]*((2.0/rhelio[i])-(1.0/a[i]))); //Geschw auf Ellipse
      //vperih[i] = sqrt(Gm[0]/a[i])*sqrt((1+Ex[i])/(1-Ex[i]));
	 

        v[i][0] = sqrt(Gm[0]/(a[i]*(1-Ex[i]*Ex[i])))*(Ex[i]*sin((nu[i]-wl[i])/180.0*Pi)*e[i][0]+(1+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi))*(C[i][1]*e[i][2]-C[i][2]*e[i][1]));
	v[i][1] = sqrt(Gm[0]/(a[i]*(1-Ex[i]*Ex[i])))*(Ex[i]*sin((nu[i]-wl[i])/180.0*Pi)*e[i][1]+(1+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi))*(C[i][2]*e[i][0]-C[i][0]*e[i][2]));
	v[i][2] = sqrt(Gm[0]/(a[i]*(1-Ex[i]*Ex[i])))*(Ex[i]*sin((nu[i]-wl[i])/180.0*Pi)*e[i][2]+(1+Ex[i]*cos((nu[i]-wl[i])/180.0*Pi))*(C[i][0]*e[i][1]-C[i][1]*e[i][0]));

	//v[i][0] = vperih[i]*(C[i][1]*e[i][2]-C[i][2]*e[i][1]);
	//v[i][1] = vperih[i]*(C[i][2]*e[i][0]-C[i][0]*e[i][2]);
	//v[i][2] = vperih[i]*(C[i][0]*e[i][1]-C[i][1]*e[i][0]);
   }
   }

   for(int i=0; i<aktZahl; i++)	//  x-
      ausgabe[i] << r[akt[i]][0]	// und y-Koordinaten
                 << "\t"		// ausgeben
                 << r[akt[i]][1] 
		 << "\t"               // 
                 << r[akt[i]][2]
		 << "\n";

   //for(long int t=0; t<=Tgesamt; t += deltaT)
   for(double t=0; t<=Tgesamt; t += deltaT)  {
       steps += 1;

       //euler(r, v, deltaT);	        // Euler-Schritt

       //eulerRichardson(r, v, deltaT);	// Euler-Richardson-Schritt


       rk4(r, v, deltaT);             // Runge-Kutta-Schritt

       if ((steps % OUTPUT)==0) {
	   for(int i=0; i<aktZahl; i++)	
	       ausgabe[i] << r[akt[i]][0]	
                 << "\t"		
                 << r[akt[i]][1] 
		 << "\t"               
                 << r[akt[i]][2]
		 << "\n";
       }
   }
  for(int i=0; i<aktZahl; i++)
     ausgabe[i].close();	// Ausgabedateien schliessen
 }
